/*
 * Parcelkeeper - support daemon for the OpenISR (R) system virtual disk
 *
 * Copyright (C) 2006-2009 Carnegie Mellon University
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation.  A copy of the GNU General Public License
 * should have been distributed along with this program in the file
 * LICENSE.GPL.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/utsname.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <linux/loop.h>
#include "nexus.h"
#include "defs.h"

static const int ignored_signals[]={SIGINT, SIGUSR1, SIGUSR2, SIGTSTP,
			SIGTTOU, 0};
static const int caught_signals[]={SIGQUIT, SIGTERM, SIGHUP, 0};

#define REQUESTS_PER_SYSCALL 64
#define MY_INTERFACE_VERSION 8
#define LOOP_UNREGISTER_TRIES 500
#if MY_INTERFACE_VERSION != NEXUS_INTERFACE_VERSION
#error This code uses a different interface version than the one defined in nexus.h
#endif

static void nexus_signal_handler(int sig)
{
	char c=sig;
	/* Race-free method of catching signals */
	(void) write(sigstate.signal_fds[1], &c, 1);
	/* The fd is set nonblocking, so if the pipe is full, the signal will
	   be lost */
}

static enum nexus_crypto crypto_to_nexus(enum cryptotype type)
{
	switch (type) {
	case CRY_AES_SHA1:
		return NEXUS_CRYPTO_AES_SHA1;
	case CRY_BLOWFISH_SHA1:
		return NEXUS_CRYPTO_BLOWFISH_SHA1;
	default:
		return NEXUS_NR_CRYPTO;
	}
}

static enum nexus_compress compress_to_nexus(enum compresstype type)
{
	switch (type) {
	case COMP_NONE:
		return NEXUS_COMPRESS_NONE;
	case COMP_ZLIB:
		return NEXUS_COMPRESS_ZLIB;
	case COMP_LZF:
		return NEXUS_COMPRESS_LZF;
	default:
		return NEXUS_NR_COMPRESS;
	}
}

static enum compresstype nexus_to_compress(enum nexus_compress type)
{
	switch (type) {
	case NEXUS_COMPRESS_NONE:
		return COMP_NONE;
	case NEXUS_COMPRESS_ZLIB:
		return COMP_ZLIB;
	case NEXUS_COMPRESS_LZF:
		return COMP_LZF;
	default:
		return COMP_UNKNOWN;
	}
}

static pk_err_t loop_bind(struct pk_state *state) {
	struct loop_info64 info;
	int i;
	int fd;

	for (i=0 ;; i++) {
		state->loopdev_name = g_strdup_printf("/dev/loop%d", i);
		fd=open(state->loopdev_name, O_RDWR);
		if (fd == -1) {
			pk_log(LOG_ERROR, "Couldn't open loop device");
			goto bad_free;
		}
		if (ioctl(fd, LOOP_GET_STATUS64, &info) && errno == ENXIO) {
			/* XXX race condition */
			if (ioctl(fd, LOOP_SET_FD, state->cache_fd)) {
				pk_log(LOG_ERROR, "Couldn't bind to loop "
							"device");
				goto bad_close;
			}
			/* This is required in order to properly configure the
			   (null) transfer function, even though it
			   shouldn't be */
			if (ioctl(fd, LOOP_GET_STATUS64, &info)) {
				pk_log(LOG_ERROR, "Couldn't get status of "
							"loop device");
				goto bad_clear;
			}
			snprintf((char*)info.lo_file_name, LO_NAME_SIZE, "%s",
						state->conf->cache_file);
			/* Set LO_FLAGS_AUTOCLEAR, which unbinds the loop
			   device after the last fd to it is closed.  This
			   flag should ensure that the device is properly
			   unbound if we crash, but is ignored on kernels
			   < 2.6.25. */
			info.lo_flags |= 4;
			if (ioctl(fd, LOOP_SET_STATUS64, &info)) {
				pk_log(LOG_ERROR, "Couldn't configure "
							"loop device");
				goto bad_clear;
			}
			break;
		}
		close(fd);
		g_free(state->loopdev_name);
	}
	state->loopdev_fd=fd;
	pk_log(LOG_INFO, "Bound to loop device %s", state->loopdev_name);
	return PK_SUCCESS;

bad_clear:
	ioctl(fd, LOOP_CLR_FD, 0);
bad_close:
	close(fd);
bad_free:
	g_free(state->loopdev_name);
	return PK_IOERR;
}

static void loop_unbind(struct pk_state *state)
{
	int i;

	/* XXX Sometimes the loop device doesn't unregister the first time.
	   For now, we retry (a lot) to try to ensure that the user isn't left
	   with a stale binding.  However, we still print the warning as a
	   debug aid. */
	for (i=0; i<LOOP_UNREGISTER_TRIES; i++) {
		if (!ioctl(state->loopdev_fd, LOOP_CLR_FD, 0)) {
			if (i > 0)
				pk_log(LOG_ERROR, "Had to try %d times to "
							"unbind loop device",
							i + 1);
			break;
		}
		usleep(10000);
	}
	if (i == LOOP_UNREGISTER_TRIES)
		pk_log(LOG_ERROR, "Couldn't unbind loop device");

	close(state->loopdev_fd);
	g_free(state->loopdev_name);
}

pk_err_t nexus_init(struct pk_state *state)
{
	struct nexus_setup setup;
	pk_err_t ret;
	unsigned u;
	gchar *revision;
	gchar *protocol;
	unsigned protocol_i;
	struct utsname utsname;

	/* Check for previous unclean shutdown of local cache */
	if (cache_test_flag(state, CA_F_DIRTY)) {
		pk_log(LOG_WARNING, "Local cache marked as dirty");
		pk_log(LOG_WARNING, "Will not run until the cache has been "
					"validated or discarded");
		return PK_BADFORMAT;
	}

	/* Check Nexus version */
	if (!g_file_test("/sys/class/openisr", G_FILE_TEST_IS_DIR)) {
		pk_log(LOG_ERROR, "kernel module not loaded");
		return PK_NOTFOUND;
	}
	if (read_sysfs_file("/sys/class/openisr/version", &protocol)) {
		pk_log(LOG_ERROR, "can't get Nexus protocol version");
		return PK_PROTOFAIL;
	}
	if (sscanf(protocol, "%u", &protocol_i) != 1) {
		pk_log(LOG_ERROR, "can't parse protocol version");
		g_free(protocol);
		return PK_PROTOFAIL;
	}
	g_free(protocol);
	if (protocol_i != MY_INTERFACE_VERSION) {
		pk_log(LOG_ERROR, "protocol mismatch: expected version "
					"%u, got version %u",
					MY_INTERFACE_VERSION, protocol_i);
		return PK_PROTOFAIL;
	}
	if (read_sysfs_file("/sys/class/openisr/revision", &revision)) {
		pk_log(LOG_ERROR, "can't get Nexus revision");
		return PK_PROTOFAIL;
	}
	pk_log(LOG_INFO, "Driver protocol %u, revision %s", protocol_i,
				revision);
	g_free(revision);

	/* Log kernel version */
	if (uname(&utsname))
		pk_log(LOG_ERROR, "Can't get kernel version");
	else
		pk_log(LOG_INFO, "%s %s (%s) on %s", utsname.sysname,
					utsname.release, utsname.version,
					utsname.machine);

	/* Create signal-passing pipe */
	if (pipe(sigstate.signal_fds)) {
		pk_log(LOG_ERROR, "couldn't create pipe");
		return PK_CALLFAIL;
	}
	/* Set it nonblocking */
	if (fcntl(sigstate.signal_fds[0], F_SETFL, O_NONBLOCK) ||
				fcntl(sigstate.signal_fds[1], F_SETFL,
				O_NONBLOCK)) {
		pk_log(LOG_ERROR, "couldn't set pipe nonblocking");
		close(sigstate.signal_fds[0]);
		close(sigstate.signal_fds[1]);
		return PK_CALLFAIL;
	}
	/* Register pipe-based signal handler */
	ret=setup_signal_handlers(nexus_signal_handler, caught_signals,
				ignored_signals);
	if (ret) {
		close(sigstate.signal_fds[0]);
		close(sigstate.signal_fds[1]);
		return ret;
	}
	if (pending_signal()) {
		/* We already got a signal under the generic signal-handling
		   system.  Exit. */
		return PK_INTERRUPT;
	}

	/* Open the device.  O_NONBLOCK ensures we never block on a read(), but
	   write() may still block */
	state->chardev_fd = open("/dev/openisrctl", O_RDWR|O_NONBLOCK);
	if (state->chardev_fd < 0) {
		if (errno == ENOENT) {
			pk_log(LOG_ERROR, "/dev/openisrctl does not exist");
			return PK_NOTFOUND;
		} else {
			pk_log(LOG_ERROR, "unable to open /dev/openisrctl");
			return PK_IOERR;
		}
	}

	/* Set the dirty flag on the local cache.  If the damaged flag is
	   already set, there's no point in forcing another check if we
	   crash. */
	if (!cache_test_flag(state, CA_F_DAMAGED)) {
		ret=cache_set_flag(state, CA_F_DIRTY);
		if (ret)
			goto bad_chardev;
	}

	/* Bind the image file to a loop device */
	ret=loop_bind(state);
	if (ret)
		goto bad_flag;

	/* Register ourselves with the device */
	memset(&setup, 0, sizeof(setup));
	snprintf((char*)setup.ident, NEXUS_MAX_DEVICE_LEN, "%s",
				state->parcel->uuid);
	snprintf((char*)setup.chunk_device, NEXUS_MAX_DEVICE_LEN, "%s",
				state->loopdev_name);
	setup.offset=state->offset >> 9;
	setup.chunksize=state->parcel->chunksize;
	setup.cachesize=(state->conf->nexus_cache << 20) /
				state->parcel->chunksize;
	setup.crypto=crypto_to_nexus(state->parcel->crypto);
	setup.compress_default=compress_to_nexus(state->conf->compress);
	for (u=0; u<8*sizeof(state->parcel->required_compress); u++)
		if (state->parcel->required_compress & (1 << u))
			setup.compress_required |= 1 << compress_to_nexus(u);
	if (setup.crypto == NEXUS_NR_CRYPTO ||
				setup.compress_default == NEXUS_NR_COMPRESS ||
				(setup.compress_required &
				(1 << NEXUS_NR_COMPRESS))) {
		/* Shouldn't happen, so we don't need a very good error
		   message */
		pk_log(LOG_ERROR, "unknown crypto or compression algorithm");
		ret = PK_IOERR;
		goto bad_loop;
	}

	if (ioctl(state->chardev_fd, NEXUS_IOC_REGISTER, &setup)) {
		pk_log(LOG_ERROR, "unable to register with Nexus: %s",
					strerror(errno));
		ret = PK_IOERR;
		goto bad_loop;
	}
	state->bdev_index=setup.index;
	pk_log(LOG_INFO, "Registered with Nexus");
	return PK_SUCCESS;

bad_loop:
	loop_unbind(state);
bad_flag:
	cache_clear_flag(state, CA_F_DIRTY);
bad_chardev:
	close(state->chardev_fd);
	return ret;
}

static void log_sysfs_value(struct pk_state *state, const char *attr)
{
	gchar *fname;
	gchar *data;

	fname = g_strdup_printf("/sys/class/openisr/openisr%c/%s",
				'a' + state->bdev_index, attr);
	if (read_sysfs_file(fname, &data)) {
		pk_log(LOG_STATS, "%s: unknown", attr);
	} else {
		pk_log(LOG_STATS, "%s: %s", attr, data);
		g_free(data);
	}
	g_free(fname);
}

void nexus_shutdown(struct pk_state *state)
{
	log_sysfs_value(state, "cache_hits");
	log_sysfs_value(state, "cache_misses");
	log_sysfs_value(state, "cache_alloc_failures");
	log_sysfs_value(state, "chunk_reads");
	log_sysfs_value(state, "chunk_writes");
	log_sysfs_value(state, "chunk_errors");
	log_sysfs_value(state, "chunk_encrypted_discards");
	log_sysfs_value(state, "whole_chunk_updates");
	log_sysfs_value(state, "sectors_read");
	log_sysfs_value(state, "sectors_written");
	log_sysfs_value(state, "compression_ratio_pct");
	pk_log(LOG_STATS, "messages_received: %u", state->request_count);

	close(state->chardev_fd);
	loop_unbind(state);
	/* We don't trust the loop driver */
	sync();
	sync();
	sync();
	if (!state->leave_dirty)
		cache_clear_flag(state, CA_F_DIRTY);
}

static void add_poll_fd(struct pk_state *state, int fd, GIOCondition cond,
			GIOFunc callback)
{
	GIOChannel *chan;
	GSource *source;

	chan = g_io_channel_unix_new(fd);
	g_io_channel_set_close_on_unref(chan, FALSE);
	source = g_io_create_watch(chan, cond);
	g_source_set_callback(source, (GSourceFunc) callback, state, NULL);
	g_source_attach(source, g_main_loop_get_context(state->nexus_loop));
	g_source_unref(source);
	g_io_channel_unref(chan);
}

static int request_is_valid(struct pk_state *state,
			const struct nexus_message *req)
{
	if (req->chunk >= state->parcel->chunks) {
		pk_log(LOG_ERROR, "Invalid chunk number %llu received "
					"from Nexus", req->chunk);
		return 0;
	}

	switch (req->type) {
	case NEXUS_MSGTYPE_GET_META:
	case NEXUS_MSGTYPE_CHUNK_ERR:
		break;
	case NEXUS_MSGTYPE_UPDATE_META:
		if (req->length > state->parcel->chunksize) {
			pk_log(LOG_ERROR, "Invalid length %u received from "
						"Nexus for chunk %llu",
						req->length, req->chunk);
			return 0;
		}
		break;
	default:
		pk_log(LOG_ERROR, "Invalid msgtype %u received from Nexus",
					req->type);
		return 0;
	}
	return 1;
}

static void chunk_error(struct pk_state *state,
			const struct nexus_message *req)
{
	const char *rw;
	enum nexus_chunk_err err;

	rw = (req->err & NEXUS_ERR_IS_WRITE) ? "writing" : "reading";
	err = req->err & ~NEXUS_ERR_IS_WRITE;
	switch (err) {
	case NEXUS_ERR_IO:
		pk_log(LOG_WARNING, "Nexus: I/O error %s chunk %llu", rw,
					req->chunk);
		break;
	case NEXUS_ERR_TAG:
		pk_log(LOG_WARNING, "Nexus: Tag check error %s chunk %llu",
					rw, req->chunk);
		log_tag_mismatch(req->expected, req->found,
					state->parcel->hashlen);
		break;
	case NEXUS_ERR_KEY:
		/* Don't log keys to the session log! */
		pk_log(LOG_WARNING, "Nexus: Key check error %s chunk %llu",
					rw, req->chunk);
		break;
	case NEXUS_ERR_HASH:
		pk_log(LOG_WARNING, "Nexus: Hashing failure %s chunk %llu",
					rw, req->chunk);
		break;
	case NEXUS_ERR_CRYPT:
		pk_log(LOG_WARNING, "Nexus: Crypto failure %s chunk %llu",
					rw, req->chunk);
		break;
	case NEXUS_ERR_COMPRESS:
		pk_log(LOG_WARNING, "Nexus: Compression failure %s chunk %llu",
					rw, req->chunk);
		break;
	default:
		pk_log(LOG_ERROR, "Unknown Nexus error (%u) %s chunk %llu",
					err, rw, req->chunk);
		break;
	}
	/* Leave the dirty bit set at shutdown, to force the local cache to
	   be checked */
	state->leave_dirty=1;
}

/* Returns true if @reply is valid */
static int process_message(struct pk_state *state,
			const struct nexus_message *request,
			struct nexus_message *reply)
{
	pk_err_t err;
	enum compresstype compress;

	if (!request_is_valid(state, request)) {
		if (request->type == NEXUS_MSGTYPE_GET_META) {
			reply->type=NEXUS_MSGTYPE_META_HARDERR;
			reply->chunk=request->chunk;
			return 1;
		}
		return 0;
	}

	switch (request->type) {
	case NEXUS_MSGTYPE_GET_META:
		reply->chunk=request->chunk;
		err=cache_get(state, request->chunk, reply->tag, reply->key,
					&compress, &reply->length);
		reply->compression=compress_to_nexus(compress);
		if (err || reply->compression == NEXUS_NR_COMPRESS)
			reply->type=NEXUS_MSGTYPE_META_HARDERR;
		else
			reply->type=NEXUS_MSGTYPE_SET_META;
		return 1;
	case NEXUS_MSGTYPE_UPDATE_META:
		/* XXX ignores errors */
		cache_update(state, request->chunk, request->tag, request->key,
					nexus_to_compress(request->compression),
					request->length);
		break;
	case NEXUS_MSGTYPE_CHUNK_ERR:
		chunk_error(state, request);
		break;
	}
	return 0;
}

static gboolean process_batch(GIOChannel *src, GIOCondition cond, void *data)
{
	struct pk_state *state = data;
	struct nexus_message requests[REQUESTS_PER_SYSCALL];
	struct nexus_message replies[REQUESTS_PER_SYSCALL];
	int i;
	int in_count;
	int out_count=0;

	in_count=read(state->chardev_fd, &requests, sizeof(requests));
	if (in_count % sizeof(requests[0]))
		pk_log(LOG_ERROR, "Short read from Nexus: %d", in_count);
	in_count /= sizeof(requests[0]);

	for (i=0; i<in_count; i++) {
		if (process_message(state, &requests[i], &replies[out_count]))
			out_count++;
		state->request_count++;
	}

	if (out_count == 0)
		return TRUE;
	out_count *= sizeof(replies[0]);
	if (write(state->chardev_fd, replies, out_count) != out_count) {
		/* XXX */
		pk_log(LOG_ERROR, "Short write to Nexus");
	}
	return TRUE;
}

static gboolean process_unregister(GIOChannel *src, GIOCondition cond,
			void *data)
{
	struct pk_state *state = data;

	if (!ioctl(state->chardev_fd, NEXUS_IOC_UNREGISTER))
		g_main_loop_quit(state->nexus_loop);
	return TRUE;
}

static gboolean process_signal(GIOChannel *src, GIOCondition cond, void *data)
{
	struct pk_state *state = data;
	char signal;

	while (read(sigstate.signal_fds[0], &signal, sizeof(signal)) > 0) {
		switch (signal) {
		case SIGQUIT:
			pk_log(LOG_INFO, "Caught SIGQUIT; shutting down "
						"immediately");
			pk_log(LOG_INFO, "Loop unregistration may fail");
			g_main_loop_quit(state->nexus_loop);
			break;
		default:
			pk_log(LOG_INFO, "Caught signal; shutdown pending");
			/* XXX multiple signals could add multiple callbacks */
			add_poll_fd(state, state->chardev_fd, G_IO_PRI,
						process_unregister);
		}
	}
	return TRUE;
}

void nexus_run(struct pk_state *state)
{
	GMainContext *ctx;

	ctx = g_main_context_new();
	state->nexus_loop = g_main_loop_new(ctx, FALSE);
	add_poll_fd(state, state->chardev_fd, G_IO_IN, process_batch);
	add_poll_fd(state, sigstate.signal_fds[0], G_IO_IN, process_signal);
	g_main_loop_run(state->nexus_loop);
	g_main_loop_unref(state->nexus_loop);
	g_main_context_unref(ctx);
}
